#include "errno.h"
#include "param.h"
#include "malloc.h"
#include "mbuf.h"
#include "protosw.h"
#include "socket.h"
#include "socketvar.h"
#include "stat.h"

#include "../net/if.h"
#include "../net/route.h"

#include "ax_pcb.h"
#include "ax.h"
#include "ax_var.h"

struct  sockaddr_ax     ui_ax = { sizeof(ui_ax), AF_AX25};

extern struct ax_addr AXADDR_ANY;

/*
 * AX.25 datagram (UI frame) implementation
 */
axd_init()
{
	printf("axd_init\n");
}

/*
 * AX.25 input routine for datagrams - UI frames carrying text (pid=0xf0)
 * m is expected to be an mbuf with an ax_hdr struct at the front
 */
axd_input(ah, m, ccb)
	register struct ax_hdr *ah;
	register struct mbuf *m;
	register struct ax_ccb *ccb;
{
	int len;

	printf("axd_input\n");

	axdstat.axds_ipackets++;

	if (ccb == 0)
		goto bad;

	/*
	 * Construct sockaddr format source address.
	 * Stuff source address and datagram in user buffer.
	 */
	ui_ax.sax_addr = ah->ax_src;
	if (sbappendaddr(&ccb->ax_socket->so_rcv,
			(struct sockaddr *)&ui_ax, m, 0) == 0) {
		axdstat.axds_fullsock++;
		goto bad;
	}
	sorwakeup(ccb->ax_socket);
	return;
bad:
	m_freem(m);
}

/*
 * Notify a axd user of an asynchronous error;
 * just wake up so that he can collect error status.
 */
axd_notify(ccb, errno)
	register struct ax_ccb *ccb;
{
	printf("axd_notify\n");

	ccb->ax_socket->so_error = errno;
	sorwakeup(ccb->ax_socket);
	sowwakeup(ccb->ax_socket);
}

axd_ctloutput(cmd, sa, ip)
	int cmd;
	struct sockaddr *sa;
	register struct ip *ip;
{
	printf("axd_ctloutput cmd=%d sa=%x ip=%x\n", cmd, sa, ip);
}


axd_ctlinput(cmd, sa, ip)
	int cmd;
	struct sockaddr *sa;
	register struct ip *ip;
{
	printf("axd_ctlinput cmd=%d sa=%x ip=%x\n", cmd, sa, ip);
#ifdef notdef
	register struct axdhdr *uh;
	extern struct in_addr zeroin_addr;
	extern u_char inetctlerrmap[];

	if ((unsigned)cmd > PRC_NCMDS || inetctlerrmap[cmd] == 0)
		return;
	if (ip) {
		uh = (struct axdhdr *)((caddr_t)ip + (ip->ip_hl << 2));
		in_pcbnotify(&axd, sa, uh->uh_dport, ip->ip_src, uh->uh_sport,
			cmd, axd_notify);
	} else
		in_pcbnotify(&axd, sa, 0, zeroin_addr, 0, cmd, axd_notify);
#endif
}

axd_output(ccb, m, addr, control)
	register struct ax_ccb *ccb;
	register struct mbuf *m;
	struct mbuf *addr, *control;
{
	int error = 0;

	printf("axd_output\n");

	if (control)
		m_freem(control);		/* XXX */

	axdstat.axds_opackets++;

	error = ax_send(ccb->hiscall, ccb->mycall, AXR_UI, AXPID_NONE,
			 0, 0, 0, 0, m, 0);

	return (error);

release:
	m_freem(m);
	return (error);
}

/*ARGSUSED*/
axd_usrreq(so, req, m, addr, control)
	struct socket *so;
	int req;
	struct mbuf *m, *addr, *control;
{
	struct ax_ccb *ccb = sotoaxccb(so);
	int error = 0;
	int s;

	printf("axd_usrreq so=%x req=%d m=%x addr=%x control=%x\n",
		so, req, m, addr, control);

	if (req == PRU_CONTROL)
		return (ax_control(so, (int)m, (caddr_t)addr,
			(struct ifnet *)control));
	if (ccb == NULL && req != PRU_ATTACH) {
		error = EINVAL;
		goto release;
	}

	/*
	 * Note: need to block axd_input while changing
	 * the axd pcb queue and/or pcb addresses.
	 */
	switch (req) {

	case PRU_ATTACH:
printf("axd_pru_attach\n");
		if (ccb) {
			error = EISCONN;
			break;
		}
		error = axs_attach(so);
		if (error)
			break;
		break;

	case PRU_DETACH:
printf("axd_pru_detach\n");
		axd_detach(ccb);
		break;

	case PRU_BIND:
printf("axd_pru_bind\n");
		s = splnet();
		error = ax_ccbbind(ccb, addr);
		splx(s);
		break;

	case PRU_LISTEN:
printf("axd_pru_listen\n");
		error = EOPNOTSUPP;
		break;

	case PRU_CONNECT:
printf("axd_pru_connect\n");
		if (ccb->state > 1) {
			error = EISCONN;
			break;
		}
		s = splnet();
		error = ax_ccbconnect(ccb, addr);
		splx(s);
		if (error == 0)
			soisconnected(so);
		break;

	case PRU_CONNECT2:
printf("axd_pru_connect2\n");
		error = EOPNOTSUPP;
		break;

	case PRU_ACCEPT:
printf("axd_pru_accept\n");
		error = EOPNOTSUPP;
		break;

	case PRU_DISCONNECT:
printf("axd_pru_disconnect\n");
		if (ccb->state < 2) {
			error = ENOTCONN;
			break;
		}
		s = splnet();
		axd_disconnect(ccb);
		splx(s);
		so->so_state &= ~SS_ISCONNECTED;		/* XXX */
		break;

	case PRU_SHUTDOWN:
printf("axd_pru_shutdown\n");
		socantsendmore(so);
		break;

	case PRU_SEND:
printf("axd_pru_send\n");
		return (axd_output(ccb, m, addr, control));

	case PRU_ABORT:
printf("axd_pru_abort\n");
		soisdisconnected(so);
		axd_detach(ccb);
		break;

	case PRU_SOCKADDR:
printf("axd_pru_sockaddr\n");
		ax_setsockaddr(ccb, addr);
		break;

	case PRU_PEERADDR:
printf("axd_pru_peeraddr\n");
		ax_setpeeraddr(ccb, addr);
		break;

	case PRU_SENSE:
printf("axd_pru_sense\n");
		/*
		 * stat: don't bother with a blocksize.
		 */
		return (0);

	case PRU_SENDOOB:
printf("axd_pru_sendoob\n");
		error =  EOPNOTSUPP;
		break;
	case PRU_FASTTIMO:
printf("axd_pru_fasttimo\n");
		error =  EOPNOTSUPP;
		break;
	case PRU_SLOWTIMO:
printf("axd_pru_slowtimo\n");
		error =  EOPNOTSUPP;
		break;
	case PRU_PROTORCV:
printf("axd_pru_protorcv\n");
		error =  EOPNOTSUPP;
		break;
	case PRU_PROTOSEND:
printf("axd_pru_protosend\n");
		error =  EOPNOTSUPP;
		break;

	case PRU_RCVD:
printf("axd_pru_rcvd\n");
		return (EOPNOTSUPP);	/* do not free mbuf's */
	case PRU_RCVOOB:
printf("axd_pru_rcvoob\n");
		return (EOPNOTSUPP);	/* do not free mbuf's */

	default:
		panic("axd_usrreq");
	}

release:
	if (control) {
		printf("axd control data unexpectedly retained\n");
		m_freem(control);
	}
	if (m)
		m_freem(m);
	return (error);
}

axd_detach(ccb)
	struct ax_ccb *ccb;
{
	printf("axd_detach\n");
#ifdef notyet
	int s = splnet();

	if (ccb == axd_last_inpcb)
		axd_last_inpcb = &axd;
	ax_ccbdetach(ccb);
	splx(s);
#endif
}

u_long	axd_sendspace = 1024*4;
u_long	axd_recvspace = 1024*4;

/*
 * Attach AX25 protocol to socket, allocating protocol control block,
 * buffer space, and entering LISTEN state if to accept connections.
 */
axd_attach(so)
	struct socket *so;
{
	register struct ax_ccb *ccb;
	int error;

	printf("axd_attach\n");

	if (so->so_snd.sb_hiwat == 0 || so->so_rcv.sb_hiwat == 0) {
		error = soreserve(so, axd_sendspace, axd_recvspace);
		if (error)
			return (error);
	}
	ccb = (struct ax_ccb *)axnewccb(AXADDR_ANY, AXADDR_ANY, AXPROTO_DGRAM);
	if (ccb == 0)
		return (ENOBUFS);
	ccb->ax_socket = so;
	so->so_pcb = (caddr_t)ccb;
	return (0);
}

/*
 * Initiate (or continue) disconnect.
 */
struct ax_ccb *
axd_disconnect(ccb)
	register struct ax_ccb *ccb;
{
	struct socket *so = ccb->ax_socket;
printf("axs_disconnect\n");
	axevent(AXE_DISC, ccb);
	soisdisconnecting(so);
	sbflush(&so->so_rcv);
	return(ccb);
}
