#include "param.h"
#include "systm.h"
#include "malloc.h"
#include "mbuf.h"
#include "domain.h"
#include "protosw.h"
#include "socket.h"
#include "errno.h"
#include "time.h"
#include "kernel.h"

#include "net/if.h"
#include "net/netisr.h"

#include "ax.h"
#include "ax_var.h"

struct ax_ccb *axnewccb();

/*
 * software interrupt handler.
 * dequeue the packet received from the interface, separate out the
 * header and the data, and pass it to ax_input for further handling
 */
axintr()
{
	register struct mbuf *m,*mh;
	register struct ax_hdr *ah;
	register u_char *rp, *ap, *ia;
	int alen,hlen,dlen,len,i,s;

next:
	/*
	 * Get next entry off input queue and separate AX header
	 *
	 */
	s = splimp();
	IF_DEQUEUE(&axintrq, m);
	splx(s);
	if (m == 0)
		return;
	len = m->m_len;

	/* it does make sense to grab another mbuf to put the parsed
	ax_header into, doesn't it?  */
	MGETHDR(mh, M_DONTWAIT, MT_DATA);
	if (mh == NULL) {	/* no mem, drop it and return */
		m_freem(m);
		return;
	}
	MH_ALIGN(mh, sizeof(struct ax_hdr));
	mh->m_len = sizeof(struct ax_hdr);
	ah = mtod(mh, struct ax_hdr *);
	bzero(ah, sizeof(struct ax_hdr));
	/*
	 * walk down the packet to find the end of the addr
	 * and rightshift the addresses while you go
	 */
	rp = ap = mtod(m, u_char *);
	ia = (u_char *)ah;	/* start of address field */
	for (alen=1; alen <= len; alen++) {
		*ia++ = *ap >> 1;
		if (*ap++ & 1)
			break;
	}
#ifdef AXDEBUG2
	printf("addrwalk = %d\n", alen);
#endif
	if (alen % 7)	/* AX.25 addresses are 7 bytes */
		printf("ax_intr: addr not multiple of 7 long: %d\n", alen);

	ah->ax_rptcnt = (alen -
		 (2 * sizeof(struct ax_addr))) / sizeof(struct ax_addr);
#ifdef AXDEBUG2
	printf("rptcnt = %d\n", ah->ax_rptcnt);
#endif
	/* get control byte */
	ah->ax_ctl = *ap++;

	if (axt_i(ah->ax_ctl) || axt_ui(ah->ax_ctl)) { /* data! */
		ah->ax_pid = *ap++;
	} else {
		ah->ax_pid = 0;
	}
	hlen = ap - rp;
	ah->ax_hlen = hlen;
	ah->ax_dlen = m->m_len - hlen;
	m_adj(m, hlen);		/* blow off real header */
	m_cat(mh, m);		/* add data to mh */

	mh->m_pkthdr.len = sizeof(struct ax_hdr) + ah->ax_dlen;
	mh->m_pkthdr.rcvif = m->m_pkthdr.rcvif;

	/* pass the rec'd packet up to ax.25 */
	ax_input(mh);
	goto next;
}

ax_input(m)
struct mbuf *m;
	{
	register struct ax_hdr *ah;
	register struct ax_ifaddr *ia;
	register struct ax_ccb *ccb;
	int i, s;

	m = m_pullup(m, sizeof(struct ax_hdr));
	if (m == 0)
		return;
	ah = mtod(m, struct ax_hdr *);

	axstat.axs_total++;

	/* separate out a few necessary bits hidden in the SSID */
	ah->ax_cmd = (ah->ax_dst.ssid & 0x40) >> 6;
	ah->ax_rbits = (ah->ax_dst.ssid & 0x30) >> 4;
	ah->ax_dst.ssid &= 0x0f;
	ah->ax_src.ssid &= 0x0f;

#ifdef AXDEBUG
	printf("Dequeued AX.25 header:\n");
	printf("ah->ax_ctl: %02x\n", ah->ax_ctl);
	printf("ah->ax_pid: %02x\n", ah->ax_pid);
	printf("ah->ax_rptcnt: %d\n", ah->ax_rptcnt);
	axaddrprint("ah->ax_dst", ah->ax_dst);
	axaddrprint("ah->ax_src", ah->ax_src);
	for (i=0; i < ah->ax_rptcnt; i++)
		axaddrprint("ah->ax_digi", ah->ax_digi[i]);
#endif
 	/*
	 * always pass a copy to the raw socket no matter what
	 */

	/*
	 * ----------------------------
	 * See if the packet is for us.
	 * ----------------------------
	 */

 	/*
	 * first, slide on down the list of repeaters.  If the first
	 * unrepeated address is us, then mark it repeated and queue
	 * it for transmission.  If the first unrepeated address isn't
	 * us, drop the packet - it's not for us.
	 */
 	for (i=0; i < ah->ax_rptcnt; i++) {
		/* gaze at last byte of each repeater address */
		if (ah->ax_digi[i].ssid & 0x40 == 0) {	/* ah, not repeated yet */
			if (axaddrmatch(ah->ax_digi[i],ax_ifaddr)) { /* us */
				ah->ax_digi[i].ssid |= 0x40; /* mark repeated */
				/*
				 * if there are more repeaters left, next
				 * hop is the next repeater.  If not, next
				 * hop is the final destination
				 */
				if (i == ah->ax_rptcnt - 1)
					ax_repeat(ah, m, ah->ax_dst);
				else
					ax_repeat(ah, m, ah->ax_digi[i+1]);
				return;
			} else
				goto bad;
		}
	}

	/*
	 * If no AX addresses have been set yet but the interfaces
	 * are receiving, can't do anything with incoming packets yet.
	 */
	if (ax_ifaddr == NULL)
		{
		printf("incoming packet but no AX.25 interface addresses set!\n");
		goto bad;
		}

	ccb = (struct ax_ccb *) 0;

	/* check for UI frames to the default broadcast address */
	if (axt_ui(ah->ax_ctl) && axaddrmatch(ah->ax_dst, ax_bcast)) {
printf("ui/defbcast match\n");
		goto ours;
	}

	/*
	 * for every CCB, match the src and dest addresses
	 */
 	for( ccb = axb.ax_ccbnext; ccb != &axb; ccb = ccb->ax_ccbnext ) {
		if ((axt_ui(ah->ax_ctl) && ccb->mode == AXPROTO_DGRAM)
		|| (!axt_ui(ah->ax_ctl) && ccb->mode != AXPROTO_DGRAM)
 		&& axaddrmatch(ccb->hiscall, ah->ax_src)
 		&& axaddrmatch(ccb->mycall, ah->ax_dst)) {
printf("ccb match\n");
 			goto ours;
		}
	}

	ccb = (struct ax_ccb *) 0;

 	/*
	 * look at the ifconfig'd addresses
	 */
#define	satosax(sa)	((struct sockaddr_ax *)(sa))
	for (ia = ax_ifaddr; ia; ia = ia->ia_next) {
		/* us explicitly? */
		if (IA_SAX(ia)->sax_family == AF_AX25
		&& axaddrmatch(IA_SAX(ia)->sax_addr, ah->ax_dst)) {
printf("iface addr match\n");
			goto ours;
		}
		/* UI and one of our broadcast addrs? */
		if (satosax(&ia->ia_broadaddr)->sax_family == AF_AX25
		&& axt_ui(ah->ax_ctl)
		&&  axaddrmatch(satosax(&ia->ia_broadaddr)->sax_addr, ah->ax_dst)) {
printf("iface bcastaddr match\n");
			goto ours;
		}
	}

	/* not for us, just dump it on the ground */
	goto bad;

ours:
	/*
	 * If the packet matched an interface address, but we don't have
	 * someone ACCEPTing or CONNECTing, there's no CCB.  Such a packet
	 * (and connection) is still valid, because they can
	 * be used to carry data for a higher-level protocol.  So we
	 * go ahead and create a CCB, but with no socket to receive the
	 * AXPID_NONE data should some arrive.  We'll handle that later.
	 * We don't need a dummy CCB for UI frames, however.
	 */
	if (ccb == 0 && !axt_ui(ah->ax_ctl))
		ccb = axnewccb(ah->ax_dst, ah->ax_src, AXPROTO_STREAM);

	if (!axt_ui(ah->ax_ctl))	/* not ui, feed it to state machine */
		if (axfsm(ah, ccb))	/* state machine says drop it */
			goto bad;
	if (!axt_i(ah->ax_ctl)		/* if it's not an I frame */
 	&&  !axt_ui(ah->ax_ctl))	/* nor a UI frame */
		goto bad;		/* we're done with it */

	/* since all the header stuff is in ah, discard it from m */
	m_adj(m, sizeof(struct ax_hdr));

	/* upcall to the enclosed protocol's input handler */
	/* (note that CCB may be a null pointer!) */
	switch (ah->ax_pid) {
case AXPID_IP:
		axipinp(ah, m, ccb);
		break;
case AXPID_ARP:
		axarpinp(ah, m, ccb);
		break;
case AXPID_NONE:
		if (axt_i(ah->ax_ctl))
			axs_input(ah, m, ccb);
		if (axt_ui(ah->ax_ctl))
			axd_input(ah, m, ccb);
		break;
case AXPID_TEXNET:
case AXPID_LQ:
case AXPID_ATALK:
case AXPID_AARP:
case AXPID_NETROM:
	default:		/* unknown protocol */
		printf("axindata: ignored frame pid=%02x\n", ah->ax_pid);
		goto bad;
	} /* switch */
	axstat.axs_delivered++;
	return;

bad:
	m_freem(m);
	return;
}

/* ARGSUSED */
ax_repeat(ah, m, dst)
struct ax_hdr *ah;
struct mbuf *m;
struct ax_addr dst;
{
	axaddrprint("ax_repeat", dst);
	m_freem(m);
}

/* ARGSUSED */
axipinp(ah, m, ccb)
struct ax_hdr *ah;
struct mbuf *m;
struct ax_ccb *ccb;
{
	int s;
	extern struct ifqueue ipintrq;

	s = splimp();
	if (IF_QFULL(&ipintrq)) {
		IF_DROP(&ipintrq);
		m_freem(m);
	} else {
		IF_ENQUEUE(&ipintrq, m);
		schednetisr(NETISR_IP);
	}
	splx(s);
}

/* ARGSUSED */
axarpinp(ah, m, ccb)
struct ax_hdr *ah;
struct mbuf *m;
struct ax_ccb *ccb;
{
	printf("axarpinp\n");
	m_freem(m);
}
