#include "param.h"
#include "systm.h"
#include "ioctl.h"
#include "malloc.h"
#include "mbuf.h"
#include "domain.h"
#include "protosw.h"
#include "socket.h"
#include "errno.h"
#include "time.h"
#include "kernel.h"
#include "socketvar.h"

#include "../net/if.h"

#include "ax.h"
#include "ax_var.h"

extern struct ax_rt axrt;

/*
 * format an ax.25 address
 */
axaddrprint(s,a)
char *s;
struct ax_addr a;
{
	int i, ssid;
	
	printf("%s ", s);
	for (i=0; i<6; i++) {
		if (a.callsign[i] == ' ')
			break;
		printf("%c", a.callsign[i]);
	}
	ssid = a.ssid & 0xf;
	printf("-%d\n", ssid);
}
/*
 * string an ax.25 address
 */
char *
axaddrstr(a)
struct ax_addr a;
{
	char *p;
	int i, ssid;
	static char buf[16];
	char tmp[8];
	
	p = buf;
	for (i=0; i<6; i++) {
		if (a.callsign[i] == ' ')
			break;
		*p++ =  a.callsign[i];
	}
	*p = 0;
	sprintf(p,"-%d", a.ssid & 0xf);
	return(buf);
}

/*
 * compare two AX.25 addresses.  Returns 1 if same, 0 if not
 */
int
axaddrmatch(a,b)
struct ax_addr a, b;
{
axaddrprint("axaddrmatch", a);
axaddrprint("         vs", b);
	if (bcmp(a.callsign, b.callsign, 6))
		return(0);
	if ((a.ssid & 0x0f) != (b.ssid & 0x0f))
		return(0);
	return(1);
}

/*
 * Fast timeout routine for processing delayed acks
 * ticks five times a second; used to update t2 timer
 */
ax_fasttimo()
{
	register struct ax_ccb *ccb;
	for (ccb = axb.ax_ccbnext; ccb != &axb; ccb = ccb->ax_ccbnext) {
		if (ccb->t2 && --ccb->t2 == 0) {
			ax_response(ccb, AXR_I);
		}
	}
}

/*
 * ticks twice a second; used to update t1 and t3 timers
 * Updates the timers in all active ccb's and
 * causes finite state machine actions if timers expire.
 */
ax_slowtimo()
{
	register struct ax_ccb *ccb;
	register struct ax_rt *rt;

	for (ccb = axb.ax_ccbnext; ccb != &axb; ccb = ccb->ax_ccbnext) {
		if (ccb->mode == AXPROTO_DGRAM)
				continue;
		if (ccb->t1 && --ccb->t1 == 0) {
			axevent(AXE_T1, ccb);
		}
		if (ccb->t3 && --ccb->t3 == 0) {
			axevent(AXE_T3, ccb);
		}
	}
	/*
	 * the only timer in dgram CCBs is the one to delete the CCB.
	 * however, deleting the CCB causes the links to shift around,
	 * and rather than figure them out, we'll just catch them next tick
	 */
	for (ccb = axb.ax_ccbnext; ccb != &axb; ccb = ccb->ax_ccbnext) {
		if (ccb->mode != AXPROTO_DGRAM)
			continue;
		if (ccb->t3 && --ccb->t3 == 0) {
			axdelccb(ccb);
			return;
		}
	}

	/*
	 * the ax25 routing table entries have to be expired too.
	 * A negative value means a permanent entry
	 */
	for (rt = axrt.rt_next; rt; rt = rt->rt_next) {
		if (rt->rt_ttl > 0)
			--rt->rt_ttl;
	}
}

int	ax_interfaces;		/* number of external internet interfaces */

/*
 * Generic ax.25 control operations (ioctl's).
 * Ifp is 0 if not an interface-specific ioctl.
 */
/* ARGSUSED */
ax_control(so, cmd, data, ifp)
	struct socket *so;
	int cmd;
	caddr_t data;
	register struct ifnet *ifp;
{
	register struct ifreq *ifr = (struct ifreq *)data;
	register struct ax_ifaddr *ia = 0;
	register struct ifaddr *ifa;
	struct ax_ifaddr *oia;
	struct ax_aliasreq *ifra = (struct ax_aliasreq *)data;
	struct mbuf *m;
	struct sockaddr_ax oldaddr;
	int error, hostIsNew, maskIsNew;
	u_long i;

	/*
	 * Find address for this interface, if it exists.
	 */
	if (ifp)
		for (ia = ax_ifaddr; ia; ia = ia->ia_next)
			if (ia->ia_ifp == ifp)
				break;

	switch (cmd) {

	case SIOCAIFADDR:
	case SIOCDIFADDR:
		printf("SIOCAIFADDR, SIOCDIFADDR\n");
		if (ifra->ifra_addr.sax_family == AF_AX25)
		if (ifra->ifra_addr.sax_family == AF_AX25)
		    for (oia = ia; ia; ia = ia->ia_next) {
			if (ia->ia_ifp == ifp  &&
			    axaddrmatch(ia->ia_addr.sax_addr,
				ifra->ifra_addr.sax_addr))
			    break;
		}
		if (cmd == SIOCDIFADDR && ia == 0)
			return (EADDRNOTAVAIL);
		/* FALLTHROUGH */
	case SIOCSIFADDR:
		printf("SIOCSIFADDR\n");
		if ((so->so_state & SS_PRIV) == 0)
			return (EPERM);
		if (ifp == 0)
			panic("in_control");
		if (ia == (struct ax_ifaddr *)0) {
			m = m_getclr(M_WAIT, MT_IFADDR);
			if (m == (struct mbuf *)NULL)
				return (ENOBUFS);
			if (ia = ax_ifaddr) {
				for ( ; ia->ia_next; ia = ia->ia_next)
					;
				ia->ia_next = mtod(m, struct ax_ifaddr *);
			} else
				ax_ifaddr = mtod(m, struct ax_ifaddr *);
			ia = mtod(m, struct ax_ifaddr *);
			if (ifa = ifp->if_addrlist) {
				for ( ; ifa->ifa_next; ifa = ifa->ifa_next)
					;
				ifa->ifa_next = (struct ifaddr *) ia;
			} else
				ifp->if_addrlist = (struct ifaddr *) ia;
			ia->ia_ifa.ifa_addr = (struct sockaddr *)&ia->ia_addr;
			ia->ia_ifa.ifa_broadaddr =
				 (struct sockaddr *)&ia->ia_broadaddr;
			if (ifp->if_flags & IFF_BROADCAST) {
				ia->ia_broadaddr.sax_len = sizeof(ia->ia_broadaddr);
				ia->ia_broadaddr.sax_family = AF_AX25;
			}
			ia->ia_ifp = ifp;
			ax_interfaces++;
		}
		break;

	case SIOCSIFBRDADDR:
		printf("SIOCSIFBRDADDR\n");
		if ((so->so_state & SS_PRIV) == 0)
			return (EPERM);
		/* FALLTHROUGH */

	case SIOCGIFADDR:
	case SIOCGIFBRDADDR:
		printf("SIOCGIFADDR,SIOCGIFBRDADDR\n");
		if (ia == (struct ax_ifaddr *)0)
			return (EADDRNOTAVAIL);
		break;

	default:
		return (EOPNOTSUPP);
		break;
	}

	switch (cmd) {

	case SIOCGIFADDR:
		*((struct sockaddr_ax *)&ifr->ifr_addr) = ia->ia_addr;
		break;

	case SIOCGIFBRDADDR:
		printf("SIOCGIFBRDADDR\n");
		if ((ifp->if_flags & IFF_BROADCAST) == 0)
			return (EINVAL);
		*((struct sockaddr_ax *)&ifr->ifr_broadaddr) = ia->ia_broadaddr;
		break;

	case SIOCSIFBRDADDR:
		printf("SIOCSIFBRDADDR\n");
		if ((ifp->if_flags & IFF_BROADCAST) == 0)
			return (EINVAL);
		ia->ia_broadaddr = *(struct sockaddr_ax *)&ifr->ifr_broadaddr;
		break;

	case SIOCSIFADDR:
		return (ax_ifinit(ifp, ia,
		    (struct sockaddr_ax *) &ifr->ifr_addr, 1));

	case SIOCAIFADDR:
		hostIsNew = 1;
		error = 0;
		if (ia->ia_addr.sax_family == AF_AX25) {
			if (ifra->ifra_addr.sax_len == 0) {
				ifra->ifra_addr = ia->ia_addr;
				hostIsNew = 0;
			} else if (axaddrmatch(ifra->ifra_addr.sax_addr,
					       ia->ia_addr.sax_addr))
				hostIsNew = 0;
		}
		if (ifra->ifra_addr.sax_family == AF_AX25 && hostIsNew)
			error = ax_ifinit(ifp, ia, &ifra->ifra_addr, 0);
		if ((ifp->if_flags & IFF_BROADCAST) &&
		    (ifra->ifra_broadaddr.sax_family == AF_AX25))
			ia->ia_broadaddr = ifra->ifra_broadaddr;
		return (error);

	case SIOCDIFADDR:
		ax_ifscrub(ifp, ia);
		if ((ifa = ifp->if_addrlist) == (struct ifaddr *)ia)
			ifp->if_addrlist = ifa->ifa_next;
		else {
			while (ifa->ifa_next &&
			       (ifa->ifa_next != (struct ifaddr *)ia))
				    ifa = ifa->ifa_next;
			if (ifa->ifa_next)
				ifa->ifa_next = ((struct ifaddr *)ia)->ifa_next;
			else
				printf("Couldn't unlink inifaddr from ifp\n");
		}
		oia = ia;
		if (oia == (ia = ax_ifaddr))
			ax_ifaddr = ia->ia_next;
		else {
			while (ia->ia_next && (ia->ia_next != oia))
				ia = ia->ia_next;
			if (ia->ia_next)
				ia->ia_next = oia->ia_next;
			else
				printf("Didn't unlink inifadr from list\n");
		}
		(void) m_free(dtom(oia));
		break;

	default:
		if (ifp == 0 || ifp->if_ioctl == 0)
			return (EOPNOTSUPP);
		return ((*ifp->if_ioctl)(ifp, cmd, data));
	}
	return (0);
}

/*
 * Delete any existing route for an interface.
 */
ax_ifscrub(ifp, ia)
	register struct ifnet *ifp;
	register struct ax_ifaddr *ia;
{
	printf("ax_ifscrub\n");
}

/*
 * Initialize an interface's internet address
 * and routing table entry.
 */
ax_ifinit(ifp, ia, sax, scrub)
	register struct ifnet *ifp;
	register struct ax_ifaddr *ia;
	struct sockaddr_ax *sax;
{
	struct sockaddr_ax oldaddr;
	int s, error, flags;

	printf("ax_ifinit\n");

	s = splimp();
	oldaddr = ia->ia_addr;
	ia->ia_addr = *sax;
	/*
	 * Give the interface a chance to initialize
	 * if this is its first address,
	 * and to validate the address if necessary.
	 */
	if (ifp->if_ioctl && (error = (*ifp->if_ioctl)(ifp, SIOCSIFADDR, ia))) {
		splx(s);
		ia->ia_addr = oldaddr;
		return (error);
	}
	splx(s);
	if (scrub) {
		ia->ia_ifa.ifa_addr = (struct sockaddr *)&oldaddr;
		ax_ifscrub(ifp, ia);
		ia->ia_ifa.ifa_addr = (struct sockaddr *)&ia->ia_addr;
	}

#ifdef notyet
	/*
	 * Add route for the network.
	 */
	flags = RTF_UP;
	if (ifp->if_flags & IFF_BROADCAST) {
		ia->ia_broadaddr.sax_addr = 
			ax_makeaddr(ia->ia_subnet, INADDR_BROADCAST);
		ia->ia_netbroadcast.s_addr =
		    htonl(ia->ia_net | (INADDR_BROADCAST &~ ia->ia_netmask));
	} else if (ifp->if_flags & IFF_LOOPBACK) {
		ia->ia_ifa.ifa_broadaddr = ia->ia_ifa.ifa_addr;
		flags |= RTF_HOST;
	}
	if ((error = rtinit(&(ia->ia_ifa), (int)RTM_ADD, flags)) == 0)
		ia->ia_flags |= IFA_ROUTE;
#endif
	return (error);
}

/*
 * Return 1 if the address might be a local broadcast address.
 */
ax_broadcast(in)
	struct ax_addr in;
{
	register struct ax_ifaddr *ia;

	printf("ax_broadcast\n");

	/* check for canonical broadcast addr */
	if (axaddrmatch(in, ax_bcast))
		     return (1);
	/*
	 * Look through the list of addresses for a match
	 * with a broadcast address.
	 */
	for (ia = ax_ifaddr; ia; ia = ia->ia_next)
	    if (ia->ia_ifp->if_flags & IFF_BROADCAST) {
		if (axaddrmatch(ia->ia_broadaddr.sax_addr, in))
		     return (1);
	    }
	return (0);
}
