#include "param.h"
#include "systm.h"
#include "malloc.h"
#include "mbuf.h"
#include "protosw.h"
#include "socket.h"
#include "socketvar.h"
#include "ioctl.h"

#include "../net/if.h"
#include "../net/route.h"
#include "../netinet/in_systm.h"
#include "../netinet/in.h"
#include "../netinet/ip.h"

#include "ax.h"
#include "ax_pcb.h"
#include "ax_var.h"

struct	ax_addr	zeroax_addr;
struct	ax_addr	AXADDR_ANY = { "any   ", 0};

/*
 * allocate another CCB.
 * Load it with initial values.
 */
struct ax_ccb *
axnewccb(me,him,mode)
struct ax_addr me, him;
int mode;
{
	struct ax_ccb *ccb;
	struct mbuf *m;
	int s;

#ifdef AXDEBUG
	printf("axnewccb\n");
	axaddrprint("   me", me);
	axaddrprint("  him", him);
#endif
	MCLALLOC((caddr_t)ccb, M_DONTWAIT);
	if (ccb == 0) {
		printf("couldn't allocate new AX_CCB\n");
		return(0);
	}
	bzero(ccb, sizeof(struct ax_ccb));
	/* stuff it in the front of the list */
	s = splnet();
	ccb->ax_ccbhead = &axb;	/* link to head of chain */
	ccb->ax_ccbprev = &axb;	/* backward link to head of chain */
	ccb->ax_ccbnext = axb.ax_ccbnext;	/* forward link in current ccb */
	if (ccb->ax_ccbnext != &axb)	/* back link in next ccb */
		ccb->ax_ccbnext->ax_ccbprev = ccb;
	axb.ax_ccbnext = ccb;		/* backward link in following ccb */
	splx(s);

	/* everything is zero except the following */
	ccb->mode = mode;
	ccb->state = 1;			/* disconnected */
	ccb->mycall = me;
	ccb->hiscall = him;
	/*
	 * we should really get the following values from the interface
	 */
	ccb->frack = 16;		/* 8 seconds */
	ccb->resptime = 2;		/* 1 second */
	ccb->keepalive = 600;		/* 5 minutes */
	ccb->paclen = 256;
	ccb->retries = 10;
	ccb->maxframe  = 1;

#ifdef AXDEBUG
	printf("new ax_ccb initialized\n");
#endif
	return(ccb);
}


/*
 * an ax.25 connection is uniquely defined by the callsign+ssid of
 * the two participants - there are no ports like TCP has, pity
 * so there's only one ccb per callsign pair.
 */
struct ax_ccb *
axfindccb(me,him,mode)
struct ax_addr me, him;
int mode;	/* raw, stream, dgram */
{
	struct ax_ccb *ccb;

#ifdef AXDEBUG
	printf("axfindccb");
	axaddrprint("   me", me);
	axaddrprint("  him", him);
#endif

	ccb = axb.ax_ccbnext;		/* start at the head of the list */
	for (;;) {
		if (ccb == &axb ) {	/* end of list */
#ifdef AXDEBUG
			printf("   no matching ax_ccb found\n");
#endif
				return(0);
		}
#ifdef AXDEBUG
		printf("   checking: mode=%d\n", ccb->mode);
		axaddrprint("   my", ccb->mycall);
		axaddrprint("  his", ccb->hiscall);
#endif
		if (mode == ccb->mode 
		&& axaddrmatch(me,ccb->mycall)
		&& axaddrmatch(him,ccb->hiscall))
			return(ccb);
		ccb = ccb->ax_ccbnext;
	}
}

/*
 * unlink and free a ccb
 */
axdelccb(ccb)
struct ax_ccb *ccb;
{
	int i;
	int s;

#ifdef AXDEBUG
	printf("axdelccb");
#endif
	if (ccb == 0)
		return;
#ifdef AXDEBUG
	axaddrprint("   me", ccb->mycall);
	axaddrprint("  him", ccb->hiscall);
#endif
	s = splnet();
	ccb->ax_ccbprev->ax_ccbnext = ccb->ax_ccbnext;
	ccb->ax_ccbnext->ax_ccbprev = ccb->ax_ccbprev;
	if (ccb->txq)
		m_freem(ccb->txq);
	if (ccb->rxq)
		m_freem(ccb->rxq);
	for (i=0; i < 8; i++)
		if (ccb->txp[i])
			m_freem(ccb->txp[i]);
	MCLFREE(ccb);
	splx(s);
}
		
ax_ccbbind(ccb, nam)
	register struct ax_ccb *ccb;
	struct mbuf *nam;
{
	register struct socket *so = ccb->ax_socket;
	register struct sockaddr_ax *sax;

	if (axaddrmatch(ccb->mycall, AXADDR_ANY) == 0)
		return (EINVAL);
	if (nam == 0)
		return(0);
	sax = mtod(nam, struct sockaddr_ax *);
	if (nam->m_len != sizeof(*sax))
		return (EINVAL);
	ccb->mycall = sax->sax_addr;
	return(0);
}

/*
 * Connect from a socket to a specified address.
 * Address must be specified in argument sax.
 * If don't have a local address for this socket yet,
 * then pick one.
 */
ax_ccbconnect(ccb, nam)
	register struct ax_ccb *ccb;
	struct mbuf *nam;
{
	struct ax_ifaddr *aa;
	struct sockaddr_ax *ifaddr;
	register struct sockaddr_ax *sax = mtod(nam, struct sockaddr_ax *);

	if (nam->m_len != sizeof (*sax))
		return (EINVAL);
	if (sax->sax_family != AF_AX25)
		return (EAFNOSUPPORT);
	if (ax_ifaddr) {
		/*
		 * If the destination address is AXADDR_ANY,
		 * use the primary local address.
		 * If the supplied address is AXADDR_BROADCAST,
		 * and the primary interface supports broadcast,
		 * choose the broadcast address for that interface.
		 */
#define	satosax(sa)	((struct sockaddr_ax *)(sa))
		if (axaddrmatch(sax->sax_addr, AXADDR_ANY))
		    sax->sax_addr = IA_SAX(ax_ifaddr)->sax_addr;
		else if (axaddrmatch(sax->sax_addr, ax_bcast) &&
		  (ax_ifaddr->ia_ifp->if_flags & IFF_BROADCAST))
		    sax->sax_addr = satosax(&ax_ifaddr->ia_broadaddr)->sax_addr;
	}
	if (axaddrmatch(ccb->mycall, AXADDR_ANY)) {
		ccb->mycall = ifaddr->sax_addr;
	}
	ccb->hiscall = sax->sax_addr;
	return (0);
}

ax_setsockaddr(ccb, nam)
	register struct ax_ccb *ccb;
	struct mbuf *nam;
{
	register struct sockaddr_ax *sax;
	
	nam->m_len = sizeof (*sax);
	sax = mtod(nam, struct sockaddr_ax *);
	bzero((caddr_t)sax, sizeof (*sax));
	sax->sax_family = AF_AX25;
	sax->sax_len = sizeof(*sax);
	sax->sax_addr = ccb->mycall;
}

ax_setpeeraddr(ccb, nam)
	struct ax_ccb *ccb;
	struct mbuf *nam;
{
	register struct sockaddr_ax *sax;
	
	nam->m_len = sizeof (*sax);
	sax = mtod(nam, struct sockaddr_ax *);
	bzero((caddr_t)sax, sizeof (*sax));
	sax->sax_family = AF_AX25;
	sax->sax_len = sizeof(*sax);
	sax->sax_addr = ccb->hiscall;
}

/*
 * Pass some notification to all connections of a protocol
 * associated with address dst.  The local address and/or port numbers
 * may be specified to limit the search.  The "usual action" will be
 * taken, depending on the ctlinput cmd.  The caller must filter any
 * cmds that are uninteresting (e.g., no error in the map).
 * Call the protocol specific routine (if any) to report
 * any errors for each matching socket.
 *
 * Must be called at splnet.
 */
ax_pcbnotify(head, dst, mycall, cmd, notify)
	struct ax_ccb *head;
	struct sockaddr *dst;
	struct ax_addr mycall;
	int cmd, (*notify)();
{
	register struct ax_ccb *ccb, *occb;
	struct ax_addr hiscall;
	int errno;
	extern u_char inetctlerrmap[];

	if ((unsigned)cmd > PRC_NCMDS || dst->sa_family != AF_AX25)
		return;
	hiscall = ((struct sockaddr_ax *)dst)->sax_addr;
	if (axaddrmatch(hiscall, AXADDR_ANY))
		return;
	errno = inetctlerrmap[cmd];
	for (ccb = head->ax_ccbnext; ccb != head;) {
		if (axaddrmatch(ccb->hiscall, hiscall) == 0 ||
		    axaddrmatch(ccb->mycall, mycall) == 0 ||
		    ccb->ax_socket == 0) {
			ccb = ccb->ax_ccbnext;
			continue;
		}
		occb = ccb;
		ccb = ccb->ax_ccbnext;
		if (notify)
			(*notify)(occb, errno);
	}
}

#ifdef notyet

struct ax_ccb *
ax_pcblookup(head, faddr, fport, laddr, lport, flags)
	struct ax_ccb *head;
	struct ax_addr faddr, laddr;
	u_short fport, lport;
	int flags;
{
	register struct ax_ccb *ccb, *match = 0;
	int matchwild = 3, wildcard;

	for (ccb = head->ccb_next; ccb != head; ccb = ccb->ccb_next) {
		if (ccb->ccb_lport != lport)
			continue;
		wildcard = 0;
		if (ccb->ccb_laddr.s_addr != AXADDR_ANY) {
			if (laddr.s_addr == AXADDR_ANY)
				wildcard++;
			else if (ccb->ccb_laddr.s_addr != laddr.s_addr)
				continue;
		} else {
			if (laddr.s_addr != AXADDR_ANY)
				wildcard++;
		}
		if (ccb->ccb_faddr.s_addr != AXADDR_ANY) {
			if (faddr.s_addr == AXADDR_ANY)
				wildcard++;
			else if (ccb->ccb_faddr.s_addr != faddr.s_addr ||
			    ccb->ccb_fport != fport)
				continue;
		} else {
			if (faddr.s_addr != AXADDR_ANY)
				wildcard++;
		}
		if (wildcard && (flags & AXPLOOKUP_WILDCARD) == 0)
			continue;
		if (wildcard < matchwild) {
			match = ccb;
			matchwild = wildcard;
			if (matchwild == 0)
				break;
		}
	}
	return (match);
}

#endif /* notyet */
