#include "param.h"
#include "socket.h"
#include "protosw.h"
#include "domain.h"
#include "mbuf.h"

#include "net/if.h"

#include "ax.h"
#include "ax_var.h"

extern	struct ax_addr ax_bcast;
u_char	ax_protox[AXPROTO_MAX];
int	axqmaxlen = IFQ_MAXLEN;

struct	ax_ifaddr *ax_ifaddr;			/* first ax25 address */

struct	sockaddr_ax axaddr = { sizeof(axaddr), AF_AX25 };

/*
 * AX.25 protocol family: 
 */
int ax_output(), ax_init(), ax_input(), ax_fasttimo(), ax_slowtimo();
int axd_input(), axd_ctlinput(), axd_ctloutput(), axd_usrreq();
int axs_input(), axs_ctlinput(), axs_ctloutput(), axs_usrreq();
int axr_input(), axr_ctloutput(), axr_usrreq();

extern	struct domain ax25domain;

struct protosw ax25sw[] = {
{ 0,		&ax25domain,	0,		0,
  0,		ax_output,	0,		0,
  0,
  ax_init,	ax_fasttimo,	ax_slowtimo,	0,
},
/* stream: connected mode, I frames */
{ SOCK_STREAM,	&ax25domain,	AXPROTO_STREAM,	PR_CONNREQUIRED|PR_WANTRCVD,
  axs_input,	0,		axs_ctlinput,	axs_ctloutput,
  axs_usrreq,
  0,		0,		0,		0,
},
/* datagram: UI frames with user supplying just the data */
{ SOCK_DGRAM,	&ax25domain,	AXPROTO_DGRAM,	PR_ATOMIC|PR_ADDR,
  axd_input,	0,		axd_ctlinput,	axd_ctloutput,
  axd_usrreq,
  0,		0,		0,		0,
},
/* datagram: UI frames with the user supplying entire frame image */
{ SOCK_RAW,	&ax25domain,	AXPROTO_RAW,	PR_ATOMIC|PR_ADDR,
  axr_input,	ax_output,	0,		axr_ctloutput,
  axr_usrreq,
  0,		0,		0,		0,
},
/* raw wildcard */
{ SOCK_RAW,	&ax25domain,	0,		PR_ATOMIC|PR_ADDR,
  ax_input,	ax_output,	0,		0,
  0,
  0,		0,		0,		0,
},
};

struct domain ax25domain =
    { AF_AX25, "hamradio", 0, 0, 0, 
      ax25sw, &ax25sw[sizeof(ax25sw)/sizeof(ax25sw[0])] };

/*
 * AX initialization: fill in IP protocol switch table.
 * All protocols not implemented in kernel go to raw AX protocol handler.
 */
ax_init()
{
	register struct protosw *pr;
	register int i;

	/* init the general broadcast address */
	bcopy("QST   ", ax_bcast.callsign, 6);
	ax_bcast.ssid = 0;

#ifdef notyet
	pr = pffindproto(PF_AX25, AXPROTO_RAW, SOCK_RAW);
	if (pr == 0)
		printf("PANIC ax_init");
	for (i = 0; i < AXPROTO_MAX; i++)
		ax_protox[i] = pr - ax25sw;
	for (pr = ax25domain.dom_protosw;
	    pr < ax25domain.dom_protoswNPROTOSW; pr++)
		if (pr->pr_domain->dom_family == PF_AX25 &&
		    pr->pr_protocol && pr->pr_protocol != AXPROTO_RAW)
			ax_protox[pr->pr_protocol] = pr - ax25sw;
#endif

	axintrq.ifq_maxlen = axqmaxlen;
	bzero(axb, sizeof axb);
	axb.ax_ccbhead = axb.ax_ccbnext = axb.ax_ccbprev = &axb;

	printf("AX.25 protocol implementation by Brian Kantor, WB6CYT\n");
}

/*
 * figure out the type of packet we just saw
 */
int
axpacketype(c)
u_char c;
{
	/* oh, we could sort these by frequency but does it really pay? */
	if (axt_i(c)	)	return(ax_pf(c)? AXPT_IPF : AXPT_I);
	if (axt_rr(c)	)	return(ax_pf(c)? AXPT_RRPF : AXPT_RR);
	if (axt_ui(c)	)	return(ax_pf(c)? AXPT_UIPF : AXPT_UI);
	if (axt_rej(c)	)	return(ax_pf(c)? AXPT_REJPF : AXPT_REJ);
	if (axt_rnr(c)	)	return(ax_pf(c)? AXPT_RNRPF : AXPT_RNR);
	if (axt_sabm(c)	)	return(ax_pf(c)? AXPT_SABMPF : AXPT_SABM);
	if (axt_ua(c)	)	return(ax_pf(c)? AXPT_UAPF : AXPT_UA);
	if (axt_disc(c)	)	return(ax_pf(c)? AXPT_DISCPF : AXPT_DISC);
	if (axt_dm(c)	)	return(ax_pf(c)? AXPT_DMPF : AXPT_DM);
	if (axt_frmr(c)	)	return(ax_pf(c)? AXPT_FRMRPF : AXPT_FRMR);
}

extern char *axresponses[];

/*
 * send a packet of some kind in response to state machine
 */
ax_response(ccb, pt)
struct ax_ccb *ccb;
int pt;
{
	u_char *p;
	register struct mbuf *m;

#ifdef AXDEBUG
	printf("ax_response: %d (%s)\n", pt, axresponses[pt]);
#endif
	switch(pt) {
case AXR_SABM:
case AXR_DISC:
		ax_send(ccb->hiscall, ccb->mycall, pt, 0, 1, 0, 0, 0,
			(struct mbuf *)0, 0);
		ccb->t1 = ccb->frack;	/* (re)start t1 timer */
		break;
case AXR_DM:
case AXR_UA:
		ax_send(ccb->hiscall, ccb->mycall, pt, 0, 0, 0, 0, 0,
			(struct mbuf *)0, 0);
		break;
case AXR_REJ:
case AXR_RNR:
case AXR_RR:
		ax_send(ccb->hiscall, ccb->mycall, pt, 0, 0, 0, ccb->vr, 0,
			(struct mbuf *)0, 0);
		break;
case AXR_RNRc:
case AXR_RRc:
		ax_send(ccb->hiscall, ccb->mycall, pt, 0, 1, 0, ccb->vr, 0,
			(struct mbuf *)0, 0);
		ccb->t1 = ccb->frack;		/* (re)start timer T1 */
		break;
case AXR_FRMR:
		/* build a FRMR info field here */
		m = m_getclr(M_WAIT, MT_DATA);
		if (m == 0) {
			printf("couldn't allocate FRMR data\n");
			return(0);
		}
		M_ALIGN(m, 3);
		p = mtod(m, u_char *);
		m->m_len = 3;
		*p++ = ccb->frmr;	/* WXYZ bits here */
		*p++ = (ccb->vr << 5) | ((ccb->rxhdr->ax_cmd & 1) << 4)
			 | (ccb->vs << 1);
		*p++ = ccb->rxhdr->ax_ctl;	/* rej frame ctl field */
		ax_send(ccb->hiscall,ccb->mycall,AXR_FRMR, 0, 0, 0, 0, 0, m, 0);
		ccb->t1 = ccb->frack;		/* (re)start timer T1 */
		break;
case AXR_I:
		/*
		 * you'd send an I frame here if you had one to send
		 */
		if (ccb->txq) {			/* any data to send? */
			axsendi(ccb,m);
		} else {
			/* no piggyback ack, have to send RR */
			if (ccb->t2 == 0) {	/* if it's time to ack, do so */
				ax_send(ccb->hiscall, ccb->mycall, AXR_RR, 0,
					0, 0, ccb->vr, 0, (struct mbuf *)0, 0);
			} else {	/* restart the del-ack timer */
				ccb->t2 = ccb->resptime;
			}
		}
		/*
		 * but if you don't, send an RR instead
		 */
		break;
	} /* switch */
}

/*
 * sends an I frame, starts timers, etc
 */
axsendi(ccb,m)
register struct ax_ccb *ccb;
register struct mbuf *m;
{
	if (ccb->vs == ((ccb->nr + ccb->maxframe) % 8))	/* window full */
		return;
	ax_send(ccb->hiscall, ccb->mycall, AXR_I, ccb->rxhdr->ax_pid, 0,
		ccb->vs, ccb->vr, 0, m, 0);
	ccb->vs = (++ccb->vs % 8);			/* bump VS */
	ccb->t1 = ccb->frack;			/* (re)start timer T1 */
	ccb->t2 = 0;				/* no delack now */
	ccb->k++;
}

/*
 * One of the AX.25 interfaces got an IP packet to send:
 * there's an IP packet in m and an IP address in dst; 
 * look dst up in the ARP table and wrap m up in an AX.25
 * packet and ship it
 */
axipout(ifp, m, dst)
	struct ifnet *ifp;
	register struct mbuf *m;
	struct sockaddr *dst;
{
/* XXX we'd use the arp table here, but it's not working yet */
struct ax_addr arpdst, metmpsrc ;
bcopy("WB6CYT", arpdst.callsign, 6);
arpdst.ssid = 0;

/* XXX we're supposed to get the src address from the interface */
bcopy("WB6CYT", metmpsrc.callsign, 6);
metmpsrc.ssid = 3;
printf("axipoutput\n");

	return(ax_send(arpdst, metmpsrc, AXR_UI, AXPID_IP, 0, 0, 0, 0, m, 0));
}
