#include "param.h"
#include "systm.h"
#include "malloc.h"
#include "mbuf.h"
#include "socket.h"
#include "socketvar.h"
#include "protosw.h"
#include "errno.h"
#include "stat.h"

#include "net/if.h"
#include "net/route.h"

#include "ax_pcb.h"
#include "ax.h"
#include "ax_var.h"

/*
 * AX25 stream protocol interface to socket abstraction.
 */

extern struct ax_addr AXADDR_ANY;

/*
 * Process a AX25 user request for AX25 tb.  If this is a send request
 * then m is the mbuf chain of send data.  If this is a timer expiration
 * (called from the software clock routine), then timertype tells which timer.
 */
/*ARGSUSED*/
axs_usrreq(so, req, m, nam, control)
	struct socket *so;
	int req;
	struct mbuf *m, *nam, *control;
{
	register struct ax_ccb *ccb;
	int s;
	int error = 0;
	int ostate;

	printf("axs_usrreq\n");

	if (req == PRU_CONTROL)
		return (ax_control(so, (int)m, (caddr_t)nam,
			(struct ifnet *)control));

	if (control && control->m_len) {
		m_freem(control);
		if (m)
			m_freem(m);
		return (EINVAL);
	}

	s = splnet();
	ccb = sotoaxccb(so);
	/*
	 * When AX25 is attached to a socket, then there will be
	 * a (struct ax_ccb) pointed at by the socket
	 */
	if (ccb == 0 && req != PRU_ATTACH) {
		splx(s);
		return (EINVAL);		/* XXX */
	}

	switch (req) {

	/*
	 * AX25 attaches to socket via PRU_ATTACH, reserving space,
	 * and a control block.
	 */
	case PRU_ATTACH:
printf("PRU_ATTACH\n");
		if (ccb) {
			error = EISCONN;
			break;
		}
		error = axs_attach(so);
		if (error)
			break;
		break;

	/*
	 * PRU_DETACH detaches the AX25 protocol from the socket.
	 */
	case PRU_DETACH:
printf("PRU_DETACH\n");
		if (ccb->state > 1)
			axevent(AXE_DISC, ccb);
		else
			axdelccb(ccb);
		break;

	/*
	 * Give the socket an address.
	 */
	case PRU_BIND:
printf("PRU_BIND\n");
		error = ax_ccbbind(ccb, nam);
		if (error)
			break;
		break;

	/*
	 * Prepare to accept connections.
	 */
	case PRU_LISTEN:
printf("PRU_LISTEN\n");
		error = ax_ccbbind(ccb, (struct mbuf *)0);	/* XXX */
		break;

	/*
	 * Initiate connection to peer.
	 * Start timers
	 */
	case PRU_CONNECT:
printf("PRU_CONNECT\n");
		error = ax_ccbconnect(ccb, nam);
		axevent(AXE_CONN, ccb);
		axsstat.axss_connattempt++;
		ccb->t1 = ccb->frack;
		ccb->t3 = ccb->keepalive;
		break;

	/*
	 * Create a AX25 connection between two sockets.
	 */
	case PRU_CONNECT2:
printf("PRU_CONNECT2\n");
		error = EOPNOTSUPP;
		break;

	/*
	 * Initiate disconnect from peer.
	 */
	case PRU_DISCONNECT:
printf("PRU_DISCONNECT\n");
		axevent(AXE_DISC, ccb);
		break;

	/*
	 * Accept a connection.  Essentially all the work is
	 * done at higher levels; just return the address
	 * of the peer, storing through addr.
	 */
	case PRU_ACCEPT: {
		struct sockaddr_ax *sax = mtod(nam, struct sockaddr_ax *);
printf("PRU_ACCEPT\n");

		nam->m_len = sizeof (struct sockaddr_ax);
		sax->sax_family = AF_AX25;
		sax->sax_len = sizeof(*sax);
		sax->sax_addr = ccb->hiscall;
		break;
		}

	/*
	 * Mark the connection as being incapable of further output.
	 */
	case PRU_SHUTDOWN:
printf("PRU_SHUTDOWN\n");
		socantsendmore(so);
		error = EOPNOTSUPP;
		break;

	/*
	 * After a receive, possibly send window update to peer.
	 */
	case PRU_RCVD:
printf("PRU_RCVD\n");
		break;

	/*
	 * Do a send by putting data in output queue
	 */
	case PRU_SEND:
printf("PRU_SEND\n");
		error = sbappend(&so->so_snd, m);
		break;

	/*
	 * Abort the AX25 connection.
	 */
	case PRU_ABORT:
printf("PRU_ABORT\n");
		axevent(AXE_DISC, ccb);
		axdelccb(ccb);
		break;

	case PRU_SENSE:
printf("PRU_SENSE\n");
		((struct stat *) m)->st_blksize = so->so_snd.sb_hiwat;
		(void) splx(s);
		return (0);

	case PRU_RCVOOB:
printf("PRU_RCVOOB\n");
		error = EINVAL;
		break;

	case PRU_SENDOOB:
printf("PRU_SENDOOB\n");
		error = EINVAL;
		break;

	case PRU_SOCKADDR:
printf("PRU_SOCKADDR\n");
		ax_setsockaddr(ccb, nam);
		break;

	case PRU_PEERADDR:
printf("PRU_PEERADDR\n");
		ax_setpeeraddr(ccb, nam);
		break;

	case PRU_SLOWTIMO:
		break;

	default:
		printf("axs_usrreq!!!\n");
	}
#ifdef notyet
	if (tp && (so->so_options & SO_DEBUG))
		tcp_trace(TA_USER, ostate, tp, (struct tcpiphdr *)0, req);
#endif
	splx(s);
	return (error);
}

axs_ctlinput(op, so, level, optname, mp)
	int op;
	struct socket *so;
	int level, optname;
	struct mbuf **mp;
{
	printf("axs_ctlinput\n");
	return(0);
}

axs_ctloutput(op, so, level, optname, mp)
	int op;
	struct socket *so;
	int level, optname;
	struct mbuf **mp;
{
	int error = 0;
	struct ax_ccb *ccb = sotoaxccb(so);
	register struct mbuf *m;

	printf("axs_ctloutput\n");

	switch (op) {

	case PRCO_SETOPT:
		m = *mp;
		switch (optname) {

		default:
			error = EINVAL;
			break;
		}
		if (m)
			(void) m_free(m);
		break;

	case PRCO_GETOPT:
		*mp = m = m_get(M_WAIT, MT_SOOPTS);
		m->m_len = sizeof(int);

		switch (optname) {
		default:
			error = EINVAL;
			break;
		}
		break;
	}
	return (error);
}

u_long	axs_sendspace = 1024*4;
u_long	axs_recvspace = 1024*4;

/*
 * Attach AX25 protocol to socket, allocating protocol control block,
 * buffer space, and entering LISTEN state if to accept connections.
 */
axs_attach(so)
	struct socket *so;
{
	register struct ax_ccb *ccb;
	int error;

	printf("axs_attach\n");

	if (so->so_snd.sb_hiwat == 0 || so->so_rcv.sb_hiwat == 0) {
		error = soreserve(so, axs_sendspace, axs_recvspace);
		if (error)
			return (error);
	}
	ccb = (struct ax_ccb *)axnewccb(AXADDR_ANY, AXADDR_ANY, AXPROTO_STREAM);
	if (ccb == 0)
		return (ENOBUFS);
	ccb->ax_socket = so;
	so->so_pcb = (caddr_t)ccb;
	return (0);
}

/*
 * Initiate (or continue) disconnect.
 */
struct ax_ccb *
axs_disconnect(ccb)
	register struct ax_ccb *ccb;
{
	struct socket *so = ccb->ax_socket;
printf("axs_disconnect\n");
	axevent(AXE_DISC, ccb);
	soisdisconnecting(so);
	sbflush(&so->so_rcv);
	return(ccb);
}

/*
 * User issued close, and wish to trail through shutdown states:
 */
struct ax_ccb *
axs_usrclosed(ccb)
	register struct ax_ccb *ccb;
{
	printf("axs_usrclosed\n");
	if (ccb && ccb->state == 1)
		soisdisconnected(ccb->ax_socket);
	return (ccb);
}

axs_input(ah, m, ccb)
	struct ax_hdr *ah;
	struct mbuf *m;
	struct ax_ccb *ccb;
{
	printf("axs_input\n");
	m_freem(m);
	return(0);
}
