/*
 * KISS ham radio AX.25 serial line interface, derived from
 * 		Serial Line interface - Rick Adams
 *
 * Brian Kantor, WB6CYT		August 1992
 */

/* from if_sl.c,v 1.11 84/10/04 12:54:47 rick Exp */

#include "ki.h"
#if NKI > 0

#include "param.h"
#include "proc.h"
#include "mbuf.h"
#include "buf.h"
#include "dkstat.h"
#include "socket.h"
#include "ioctl.h"
#include "file.h"
#include "tty.h"
#include "kernel.h"
#include "conf.h"

#include "machine/mtpr.h"

#include "../net/if.h"
#include "../net/if_types.h"
#include "../net/netisr.h"

#include "ax.h"

#include "if_ki.h"

static struct mbuf *ki_btom();

/* longest packet is 1 ctlbyte + 72 bytes ax.25 hdr + KIMTU bytes data */
#define	KIMTU		256	/* official MTU (max data size) */

#define	KIBUFSIZE	MCLBYTES	/* real MTU */

#define	CLISTRESERVE	1024	/* Can't let clists get too low */

struct ki_softc ki_softc[NKI];

#define FRAME_END	 	0xc0		/* Frame End */
#define FRAME_ESCAPE		0xdb		/* Frame Esc */
#define TRANS_FRAME_END	 	0xdc		/* transposed frame end */
#define TRANS_FRAME_ESCAPE 	0xdd		/* transposed frame esc */

#define t_kc T_LINEP

int kioutput(), kiioctl(), ttrstrt();

extern struct timeval time;

extern struct ifqueue axintrq;

/*
 * Called from boot code to establish ki interfaces.
 */
kiattach()
{
	register struct ki_softc *kc;
	register int i = 0;

	for (kc = ki_softc; i < NKI; kc++) {
		kc->kc_if.if_name = "ki";
		kc->kc_if.if_unit = i++;
		kc->kc_if.if_mtu = KIBUFSIZE;
		kc->kc_if.if_flags = (IFF_PROMISC | IFF_ALLMULTI |
			IFF_NOTRAILERS | IFF_BROADCAST | IFF_SIMPLEX);
		kc->kc_if.if_type = IFT_KISS;
		kc->kc_if.if_addrlen = 7;
		kc->kc_if.if_ioctl = kiioctl;
		kc->kc_if.if_output = kioutput;
		kc->kc_if.if_snd.ifq_maxlen = 50;
		if_attach(&kc->kc_if);
	}
}

static int
kiinit(kc)
	register struct ki_softc *kc;
{
	u_char *p;

	if (kc->kc_buf == (u_char *) 0) {
		MCLALLOC((u_char *)p, M_WAIT);
		if (!p) {
			printf("ki%d: can't allocate buffer\n", kc - ki_softc);
			kc->kc_if.if_flags &= ~IFF_UP;
			return (0);
		}
		kc->kc_buf = p;
	}
	kc->kc_mp = kc->kc_buf;
	kc->kc_ep = kc->kc_buf + KIBUFSIZE;
	return (1);
}

/*
 * Line specific open routine.
 * Attach the given tty to the first available ki unit.
 */
/* ARGSUSED */
kiopen(dev, tp)
	dev_t dev;
	register struct tty *tp;
{
	struct proc *p = curproc;		/* XXX */
	register struct ki_softc *kc;
	register int nki;
	int error;

	if (error = suser(p->p_ucred, &p->p_acflag))
		return (error);

	if (tp->t_line == KISSDISC)
		return (0);

	for (nki = NKI, kc = ki_softc; --nki >= 0; kc++)
		if (kc->kc_ttyp == NULL) {
			if (kiinit(kc) == 0)
				return (ENOBUFS);
			tp->t_kc = (caddr_t)kc;
			kc->kc_ttyp = tp;
			kc->kc_if.if_baudrate = tp->t_ospeed;
			kc->kc_if.if_flags |= IFF_RUNNING | IFF_UP ;
			ttyflush(tp, FREAD | FWRITE);
			printf("ki%d attached\n", kc->kc_if.if_unit);
			return (0);
		}
	return (ENXIO);
}

/*
 * Line specific close routine.
 * Detach the tty from the ki unit.
 * Mimics part of ttyclose().
 */
kiclose(tp)
	struct tty *tp;
{
	register struct ki_softc *kc;
	int s;

	ttywflush(tp);
	s = splimp();		/* actually, max(spltty, splnet) */
	tp->t_line = 0;
	kc = (struct ki_softc *)tp->t_kc;
	if (kc != NULL) {
		printf("ki%d detached\n", kc->kc_if.if_unit);
		if_down(&kc->kc_if);
		kc->kc_ttyp = NULL;
		tp->t_kc = NULL;
		MCLFREE(kc->kc_buf);
		kc->kc_ep = 0;
		kc->kc_mp = 0;
		kc->kc_buf = 0;
	}
	splx(s);
}

/*
 * Line specific (tty) ioctl routine.
 * Provide a way to get the ki unit number.
 */
/* ARGSUSED */
kitioctl(tp, cmd, data, flag)
	struct tty *tp;
	caddr_t data;
{
	struct ki_softc *kc = (struct ki_softc *)tp->t_kc;
	int s;

#ifdef KIDEBUGO
	printf("kitioctl()\n");
#endif

	switch (cmd) {
	case KIIOCGUNIT:
		*(int *)data = kc->kc_if.if_unit;
		break;

	case KIIOCGFLAGS:
		*(int *)data = kc->kc_flags;
		break;

	case KIIOCSFLAGS:
#define	SC_MASK	0xffff
		s = splimp();
		kc->kc_flags =
		    (kc->kc_flags &~ SC_MASK) | ((*(int *)data) & SC_MASK);
		splx(s);
		break;

	default:
		return (-1);
	}
	return (0);
}

/*
 * Queue a packet.  Start transmission if not active.
 */
kioutput(ifp, m, dst)
	struct ifnet *ifp;
	register struct mbuf *m;
	struct sockaddr *dst;
{
	register struct ki_softc *kc = &ki_softc[ifp->if_unit];
	register struct ifqueue *ifq;
	int s;

#ifdef KIDEBUGO
	printf("kioutput()\n");
#endif

printf("kioutput: dst.af=%d\n", dst->sa_family);

	/*
	 * if the address family is already AX.25, we've got a packet in m
	 * if it's INET, we've got to wrap the IP into AX.25
	 */
	if (dst->sa_family == AF_INET)
		return(axipout(ifp, m, dst));

	if (dst->sa_family != AF_AX25) {
		printf("ki%d: af%d not supported\n", kc->kc_if.if_unit,
			dst->sa_family);
		m_freem(m);
		return (EAFNOSUPPORT);
	}

	if (kc->kc_ttyp == NULL) {
		m_freem(m);
		return (ENETDOWN);	/* sort of */
	}
	if ((kc->kc_ttyp->t_state & TS_CARR_ON) == 0) {
		m_freem(m);
		return (EHOSTUNREACH);
	}
	ifq = &kc->kc_if.if_snd;
	s = splimp();
	if (IF_QFULL(ifq)) {
		IF_DROP(ifq);
		m_freem(m);
		splx(s);
		kc->kc_if.if_oerrors++;
		printf("ki_output: no room to queue packet\n");
		return (ENOBUFS);
	}

	IF_ENQUEUE(ifq, m);
	kc->kc_if.if_lastchange = time;
	if (RB_LEN(&kc->kc_ttyp->t_out) == 0) {
		kistart(kc->kc_ttyp);
	}
	splx(s);
	return(0);
}

/*
 * Start output on interface.  Get another datagram
 * to send from the interface queue and map it to
 * the interface before starting output.
 */
kistart(tp)
	register struct tty *tp;
{
	register struct ki_softc *kc = (struct ki_softc *)tp->t_kc;
	register struct mbuf *m;
	register struct mbuf *m2;
	register u_char *bp, *cp, *ep;
	int cc;
	int s;
	int i;

#ifdef KIDEBUGO
	printf("kistart()\n");
#endif

	for (;;) {
		/*
		 * If there is more in the output queue, just send it now.
		 * We are being called in lieu of ttstart and must do what
		 * it would.
		 */
		if (RB_LEN(&tp->t_out) != 0) {
			(*tp->t_oproc)(tp);
		}

		/*
		 * This happens briefly when the line shuts down.
		 */
		if (kc == NULL)
			return;

		/*
		 * Get a packet and send it to the interface.
		 */
		s = splimp();
		IF_DEQUEUE(&kc->kc_if.if_snd, m);
		splx(s);
		if (m == NULL)
			return;

		kc->kc_if.if_lastchange = time;
		/*
		 * If system is getting low on clists, just flush our
		 * output queue (if the stuff was important, it'll get
		 * retransmitted).
		 */
		if (RBSZ - RB_LEN(&tp->t_out) < KIMTU) {
			m_freem(m);
			kc->kc_if.if_collisions++;
			printf("kistart: clist shortage flushed output queue\n");
			continue;
		}

		/*
		 * A FRAME_END will start up a new packet, and thus
		 * will flush any accumulated garbage in the rx.
		 */
		(void) putc(FRAME_END, &tp->t_out);
		++kc->kc_bytessent;

		/* KISS data frames start with a zero command byte */
		(void) putc(0, &tp->t_out);
		++kc->kc_bytessent;

		while (m) {
			cp = mtod(m, u_char *);
			ep = cp + m->m_len;
			while (cp < ep) {
				/*
				 * Find out how many bytes in the string we can
				 * handle without doing something special.
				 */
				bp = cp;
				/* scan for special characters */
				while (cp < ep) {
					if (*cp == FRAME_ESCAPE || *cp == FRAME_END)
						break;
					cp++;
				}

				/*
				 * Put cc chars at once into the tty output queue.
				 */
				while(cp > bp && (cc=RB_CONTIGPUT(&tp->t_out))) {
					cc = min(cc, cp - bp);
					bcopy((char *)bp, tp->t_out.rb_tl, cc);
					tp->t_out.rb_tl = RB_ROLLOVER(&tp->t_out,
						 tp->t_out.rb_tl + cc);
					kc->kc_bytessent += cc;
					bp += cc;
				}
				/*
				 * If there are characters left in the mbuf,
				 * the first one must be special..
				 * Put it out in a different form.
				 */
				if (cp < ep) {
					if (putc(FRAME_ESCAPE, &tp->t_out)) {
						printf("kioutput: can't fit FE\n");
						break;
					}
					if (putc(*cp++ == FRAME_ESCAPE ?
					   TRANS_FRAME_ESCAPE : TRANS_FRAME_END,
					   &tp->t_out)) {
						printf("kioutput: can't fit TFE\n");
						break;
					}
					kc->kc_bytessent += 2;
				}
			}
			MFREE(m, m2);
			m = m2;
		}

		if (putc(FRAME_END, &tp->t_out)) {
			printf("kioutput: can't fit FEND\n");
			break;
		}
		++kc->kc_bytessent;
		kc->kc_if.if_opackets++;
		kc->kc_if.if_obytes = kc->kc_bytessent;
	}
}


/*
 * tty interface receiver interrupt.  Called once per character
 */
kiinput(c, tp)
	register int c;
	register struct tty *tp;
{
	register struct ki_softc *kc;
	register struct mbuf *m;
	register int len;		/* packet length total */
	register int i;
	int s;

	tk_nin++;
	kc = (struct ki_softc *)tp->t_kc;
	if (kc == NULL)
		return;
	if (!(tp->t_state&TS_CARR_ON))	/* XXX */
		return;

	++kc->kc_bytesrcvd;
	++kc->kc_if.if_ibytes;
	c &= 0xff;			/* XXX */

	/* is it a special (protocol-significant) character? */
	switch (c) {

	case TRANS_FRAME_ESCAPE:
		if (kc->kc_escape)
			c = FRAME_ESCAPE;
		break;

	case TRANS_FRAME_END:
		if (kc->kc_escape)
			c = FRAME_END;
		break;

	case FRAME_ESCAPE:
		kc->kc_escape = 1;
		return;

	case FRAME_END:
		len = kc->kc_mp - kc->kc_buf;
		/* less than min length packet - ignore */
		if (len < 16)
			goto newpack;
#ifdef KIDEBUGI
		printf("kiinput frame of %d bytes\n", len);
		for (i=0; i<len; i++)
			printf(" %02x", *(kc->kc_buf + i));
		printf("\n");
#endif
		/* KISS data frames ALWAYS start with a zero flag byte */
		if ((c = *kc->kc_buf) != 0) {
			printf("%s%d rx martian len %d, %02x %02x\n",
				kc->kc_if.if_name, kc->kc_if.if_unit,
				len, c, *(kc->kc_buf+1));
			goto error;
		}

		if (len > MCLBYTES) {
			printf("kiinput: packet too long: %d\n", len);
			goto error;
			}

		m = ki_btom(kc, len);
		if (m == NULL)
			goto error;

		m_adj(m, 1);	/* discard 0 kiss-control byte */

		kc->kc_if.if_ipackets++;
		kc->kc_if.if_lastchange = time;

		/* pass the rec'd packet up to ax.25 */
		s = splimp();
		if (IF_QFULL(&axintrq)) {
			IF_DROP(&axintrq);
			kc->kc_if.if_ierrors++;
			kc->kc_if.if_iqdrops++;
			m_freem(m);
			printf("ki%d: mbuf dropped - axintrq full\n",
				 kc - ki_softc);
		} else {
			IF_ENQUEUE(&axintrq, m);
			schednetisr(NETISR_AX25);
		}
		splx(s);
		goto newpack;
	} /* end switch */

	/* char isn't special, append to buffer and return */
	if (kc->kc_mp < kc->kc_ep) {	/* note we silently truncate to buflen */
		*kc->kc_mp++ = c;
		kc->kc_escape = 0;
		return;
	}

error:
	kc->kc_if.if_ierrors++;

newpack:
	kc->kc_mp = kc->kc_buf = kc->kc_ep - KIBUFSIZE;
	kc->kc_escape = 0;
	return;
}

/*
 * Copy data buffer to mbuf chain; add ifnet pointer.
 */
static struct mbuf *
ki_btom(kc, len)
	register struct ki_softc *kc;
	register int len;
{
	register struct mbuf *m;

	MGETHDR(m, M_DONTWAIT, MT_DATA);
	if (m == NULL)
		return (NULL);

	/*
	 * If we have more than MHLEN bytes, it's cheaper to
	 * queue the cluster we just filled & allocate a new one
	 * for the input buffer.  Otherwise, fill the mbuf we
	 * allocated above.  Note that code in the input routine
	 * guarantees that packet will fit in a cluster.
	 */
	if (len >= MHLEN) {
		MCLGET(m, M_DONTWAIT);
		if ((m->m_flags & M_EXT) == 0) {
			/*
			 * we couldn't get a cluster - if memory's this
			 * low, it's time to start dropping packets.
			 */
			(void) m_free(m);
			return (NULL);
		}
		kc->kc_ep = mtod(m, u_char *) + KIBUFSIZE;
		m->m_data = (caddr_t)kc->kc_buf;
		m->m_ext.ext_buf = (caddr_t)((int)kc->kc_buf &~ MCLOFSET);
	} else
		bcopy((caddr_t)kc->kc_buf, mtod(m, caddr_t), len);

	m->m_len = len;
	m->m_pkthdr.len = len;
	m->m_pkthdr.rcvif = &kc->kc_if;
	return (m);
}

/*
 * Process an ioctl request.
 */
kiioctl(ifp, cmd, data)
	register struct ifnet *ifp;
	int cmd;
	caddr_t data;
{
	register struct ifaddr *ifa = (struct ifaddr *)data;
	struct ki_softc *kc = &ki_softc[ifp->if_unit];
	struct ifreq *ifr = (struct ifreq *)data;
	int s, error = 0;

	s = splimp();

	switch (cmd) {

	case SIOCSIFADDR:
	case SIOCSIFBRDADDR:
		ifp->if_flags |= IFF_UP;

		switch (ifa->ifa_addr->sa_family) {
		case AF_INET:
			break;
#ifdef AX
		case AF_AX25:
#ifdef notyet
			kc->kc_if.ifa_addr = IA_SAX(ifa)->sax_addr;
#endif /* notyet */
			break;
#endif
		default:
			error = EAFNOSUPPORT;
			break;
		}
		break;

	case SIOCSIFFLAGS:
		if ((ifp->if_flags & IFF_UP) == 0 &&
		    ifp->if_flags & IFF_RUNNING) {
			ifp->if_flags &= ~IFF_RUNNING;
		}
		break;

	default:
		error = EINVAL;
	}
	splx(s);
	return (error);
}
#endif /* NKI */
