/*
 * Copyright (c) 1983, 1988 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static char sccsid[] = "@(#)ax25.c	5.15 (Berkeley) 6/18/90";
#endif /* not lint */

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/mbuf.h>
#include <sys/protosw.h>

#include <net/if.h>

#include <netham/ax.h>
#include <netham/ax_var.h>
#include <netham/ax_debug.h>
#include <netham/ax_pcb.h>

#include <netdb.h>

#include <stdio.h>
#include <string.h>

struct	ax_ccb axb;
struct	socket sockb;

extern	int Aflag;
extern	int aflag;
extern	int nflag;
extern	char *plural();

char *axaddrstr();

/*
 * Print a summary of connections related to the AX25
 * protocol.  Also give state of connection.
 * Listening processes (aflag) are suppressed unless the
 * -a (all) flag is specified.
 */
axprotopr(off, name)
	off_t off;
	char *name;
{
	struct ax_ccb cb;
	register struct ax_ccb *prev, *next;
	static int first = 1;
	int i;

	if (off == 0)
		return;
	if (kvm_read(off, (char *)&cb, sizeof(struct ax_ccb)) < 0) {
		perror("kvm_read");
		return;
	}
	axb = cb;
	prev = (struct ax_ccb *)off;
	/* empty list? */
	if (axb.ax_ccbnext == (struct ax_ccb *)off)
		return;
	while (axb.ax_ccbnext != (struct ax_ccb *)off) {
		next = axb.ax_ccbnext;
		if (kvm_read((off_t)next, (char *)&axb, sizeof(axb)) < 0) {
			perror("kvm_read");
			return;
		}
		if (axb.ax_ccbprev != prev) {
			printf("???\n");
/* XXX			break;	*/
		}
#ifdef notyet
		/* skip server CCBs */
		if (!aflag && axb.ax_hiscall == AXADDR_ANY) {
			prev = next;
			continue;
		}
#endif
		if (kvm_read((off_t)axb.ax_socket, (char *)&sockb, sizeof (sockb)) < 0) {
			perror("kvm_read");
			return;
		}
		if (first) {
			printf("Active AX25 connections");
			if (aflag)
				printf(" (including servers)");
			putchar('\n');
			if (Aflag)
				printf("%-8.8s ", "CCB");
			printf("%-5.5s %-6.6s %-6.6s  %-8.8s %-8.8s %-8.8s %s\n",
				"Proto", "Recv-Q", "Send-Q",
				"My_Call", "His_Call", "(mode)", "(state)");
			first = 0;
		}
		if (Aflag)
			printf("%8x ", next);
		printf("%-5.5s %6d %6d ", name, sockb.so_rcv.sb_cc,
			sockb.so_snd.sb_cc);
		printf(" %-8.8s", axaddrstr(axb.mycall));
		printf(" %-8.8s", axaddrstr(axb.hiscall));
		printf(" %-8.8s", axmodes[axb.mode]);
		printf(" %s", axstates[axb.state]);
		putchar('\n');
		prev = next;
	}
}

/*
 * Dump connected-mode AX.25 statistics structure.
 */
axs_stats(off, name)
	off_t off;
	char *name;
{
	struct axsstat axsstat;

puts("axsstats");

	if (off == 0)
		return;
	printf ("%s:\n", name);
	kvm_read(off, (char *)&axsstat, sizeof (axsstat));

#define	p(f, m)		printf(m, axsstat.f, plural(axsstat.f))
#define	p2(f1, f2, m)	printf(m, axsstat.f1, plural(axsstat.f1), axsstat.f2, plural(axsstat.f2))
  
	p(axss_sndtotal, "\t%d packet%s sent\n");
	p2(axss_sndpack,axss_sndbyte,
		"\t\t%d data packet%s (%d byte%s)\n");
	p2(axss_sndrexmitpack, axss_sndrexmitbyte,
		"\t\t%d data packet%s (%d byte%s) retransmitted\n");
	p2(axss_sndacks, axss_delack,
		"\t\t%d ack-only packet%s (%d delayed)\n");
	p(axss_sndurg, "\t\t%d URG only packet%s\n");
	p(axss_sndprobe, "\t\t%d window probe packet%s\n");
	p(axss_sndwinup, "\t\t%d window update packet%s\n");
	p(axss_sndctrl, "\t\t%d control packet%s\n");
	p(axss_rcvtotal, "\t%d packet%s received\n");
	p2(axss_rcvackpack, axss_rcvackbyte, "\t\t%d ack%s (for %d byte%s)\n");
	p(axss_rcvdupack, "\t\t%d duplicate ack%s\n");
	p(axss_rcvacktoomuch, "\t\t%d ack%s for unsent data\n");
	p2(axss_rcvpack, axss_rcvbyte,
		"\t\t%d packet%s (%d byte%s) received in-sequence\n");
	p2(axss_rcvduppack, axss_rcvdupbyte,
		"\t\t%d completely duplicate packet%s (%d byte%s)\n");
	p2(axss_rcvpartduppack, axss_rcvpartdupbyte,
		"\t\t%d packet%s with some dup. data (%d byte%s duped)\n");
	p2(axss_rcvoopack, axss_rcvoobyte,
		"\t\t%d out-of-order packet%s (%d byte%s)\n");
	p2(axss_rcvpackafterwin, axss_rcvbyteafterwin,
		"\t\t%d packet%s (%d byte%s) of data after window\n");
	p(axss_rcvwinprobe, "\t\t%d window probe%s\n");
	p(axss_rcvwinupd, "\t\t%d window update packet%s\n");
	p(axss_rcvafterclose, "\t\t%d packet%s received after close\n");
	p(axss_rcvbadsum, "\t\t%d discarded for bad checksum%s\n");
	p(axss_rcvbadoff, "\t\t%d discarded for bad header offset field%s\n");
	p(axss_rcvshort, "\t\t%d discarded because packet too short\n");
	p(axss_connattempt, "\t%d connection request%s\n");
	p(axss_accepts, "\t%d connection accept%s\n");
	p(axss_connects, "\t%d connection%s established (including accepts)\n");
	p2(axss_closed, axss_drops,
		"\t%d connection%s closed (including %d drop%s)\n");
	p(axss_conndrops, "\t%d embryonic connection%s dropped\n");
	p2(axss_rttupdated, axss_segstimed,
		"\t%d segment%s updated rtt (of %d attempt%s)\n");
	p(axss_rexmttimeo, "\t%d retransmit timeout%s\n");
	p(axss_timeoutdrop, "\t\t%d connection%s dropped by rexmit timeout\n");
	p(axss_persisttimeo, "\t%d persist timeout%s\n");
	p(axss_keeptimeo, "\t%d keepalive timeout%s\n");
	p(axss_keepprobe, "\t\t%d keepalive probe%s sent\n");
	p(axss_keepdrops, "\t\t%d connection%s dropped by keepalive\n");
#undef p
#undef p2
}

/*
 * Dump AX.25 datagram (UI) statistics structure.
 */
axd_stats(off, name)
	off_t off;
	char *name;
{
	struct axdstat axdstat;

puts("axdstats");

	if (off == 0)
		return;
	kvm_read(off, (char *)&axdstat, sizeof (axdstat));
	printf("%s:\n\t%u incomplete header%s\n", name,
		axdstat.axds_hdrops, plural(axdstat.axds_hdrops));
	printf("\t%u bad data length field%s\n",
		axdstat.axds_badlen, plural(axdstat.axds_badlen));
	printf("\t%u bad checksum%s\n",
		axdstat.axds_badsum, plural(axdstat.axds_badsum));
}

/*
 * Dump AX25 statistics structure.
 */
ax_stats(off, name)
	off_t off;
	char *name;
{
	struct axstat axstat;

puts("axstats");

	if (off == 0)
		return;
	kvm_read(off, (char *)&axstat, sizeof (axstat));
	printf("%s:\n\t%u total packets received\n", name,
		axstat.axs_total);
	printf("\t%u bad header checksum%s\n",
		axstat.axs_badsum, plural(axstat.axs_badsum));
	printf("\t%u with size smaller than minimum\n", axstat.axs_tooshort);
	printf("\t%u with data size < data length\n", axstat.axs_toosmall);
	printf("\t%u with header length < data size\n", axstat.axs_badhlen);
	printf("\t%u with data length < header length\n", axstat.axs_badlen);
	printf("\t%u fragment%s received\n",
		axstat.axs_fragments, plural(axstat.axs_fragments));
	printf("\t%u fragment%s dropped (dup or out of space)\n",
		axstat.axs_fragdropped, plural(axstat.axs_fragdropped));
	printf("\t%u fragment%s dropped after timeout\n",
		axstat.axs_fragtimeout, plural(axstat.axs_fragtimeout));
	printf("\t%u packet%s forwarded\n",
		axstat.axs_forward, plural(axstat.axs_forward));
	printf("\t%u packet%s not forwardable\n",
		axstat.axs_cantforward, plural(axstat.axs_cantforward));
	printf("\t%u redirect%s sent\n",
		axstat.axs_redirectsent, plural(axstat.axs_redirectsent));
}

/*
 * string an ax.25 address
 */
char *
axaddrstr(a)
struct ax_addr a;
{
	char *p;
	int i, ssid;
	static char buf[16];
	char tmp[8];
	
	p = buf;
	for (i=0; i<6; i++) {
		if (a.callsign[i] == ' ')
			break;
		*p++ =  a.callsign[i];
	}
	*p = 0;
	sprintf(p,"-%d", a.ssid & 0xf);
	return(buf);
}
